(***************************************************************************
 *
 * Abstract Syntax Tree Implementation
 *
 * Author: Gerald Baumgartner
 * Date:   18 October 2001
 * Date:   9 April 2010
 *
 ***************************************************************************)

structure AbSyn : AB_SYN =
  struct
    type iSet = IndexSetSeq.iSet
    type iSetSeq = IndexSetSeq.iSetSeq
    type expTree = ExpTree.expTree
    type fusionTree = FusionTree.fusionTree

    datatype arithOp = Add | Sub | Mul | Div

    datatype abSyn = Null
		   | IntLit of IntInf.int
		   | Var of string
		   | Fun of string * string list
		   | Array of string * string list
		   | BinOp of arithOp * abSyn * abSyn
		   | Assign of abSyn * abSyn
		   | OpAssign of arithOp * abSyn * abSyn
		   | For of iSet * abSyn
		   | Seq of abSyn list
		   | Write of string * string list * abSyn
		   | Return of abSyn
		   | VarDecl of string * int list
		   | FunDecl of string * string * int list
		   | TempDecl of string * int list
		   | ResultDecl of string * int list
		   | Function of string * string * abSyn
		   | Alloc of string * int list
		   | Free of string * int list

    exception MalformedFusionTree of iSetSeq * (iSetSeq*IntInf.int*abSyn) list

    local
      fun toStringOp Add = "+"
        | toStringOp Sub = "-"
        | toStringOp Mul = "*"
        | toStringOp Div = "/"

      fun expSize (IntLit _) = 1
        | expSize (Var _) = 1
        | expSize (Fun _) = 1
        | expSize (Array _) = 1
        | expSize (BinOp (_, l, r)) = expSize l + expSize r
        | expSize _ = 0

      (* returns a string of n spaces *)
      fun indent n = if n = 0 then "" else " " ^ indent (n - 1)
    in
      local

	fun toStringArgs nil = ""
	  | toStringArgs [h] = h
	  | toStringArgs (h::t) = h ^ "," ^ toStringArgs t

	fun toString' (Null, _) = ""
	  | toString' (IntLit i, _) = IntInf.toString i
	  | toString' (Var v, _) = v
	  | toString' (Fun (f, i), _) = f ^ "(" ^ toStringArgs i ^ ")"
	  | toString' (Array (a, i), _) = a ^ "[" ^ toStringArgs i ^ "]"
	  | toString' (t as BinOp (O, l, r), n) =
	      let val s = expSize t
	      in
		(if O = Add orelse O = Sub then "(" else "") ^ toString' (l, n)
		  ^ (if s <= 5 then " " else "\n" ^ indent (n+2))
		  ^ toStringOp O ^ " " ^ toString' (r, n)
		  ^ (if O = Add orelse O = Sub then ")" else "")
	      end
	  | toString' (Assign (l, r), n) =
	      indent n ^ toString' (l, n) ^ " = " ^ toString' (r, n) ^ "\n"
	  | toString' (OpAssign (O, l, r), n) =
	      indent n ^ toString' (l, n) ^ " " ^ toStringOp O ^ "= "
		^ toString' (r, n) ^ "\n"
	  | toString' (For (nil, b), n) = toString' (b, n)
	  | toString' (For (i, b), n) =
	      indent n ^ "for " ^ toStringArgs i ^ "\n"
		^ toString' (b, n+2)
	  | toString' (Seq l, n) =
	      concat (map (fn t => toString' (t, n)) l)
	  | toString' (Write (_, _, l), n) =
	      indent n ^ "write " ^ toString' (l, n+2) ^ "\n"
	  | toString' (Return l, n) =
	      indent n ^ "return " ^ toString' (l, n+2) ^ "\n"
	  | toString' (VarDecl _, _) = ""
	  | toString' (FunDecl _, _) = ""
	  | toString' (TempDecl _, _) = ""
	  | toString' (ResultDecl _, _) = ""
	  | toString' (Function (_, _, b), n) = "\n" ^ toString' (b, n)
	  | toString' (Alloc (_, nil), _) = ""
	  | toString' (Alloc (a, i), n) =
	      indent n ^ "allocate " ^ a ^ ": " ^
	      Int.toString (foldr op* 1 i) ^ "\n"
	  | toString' (Free (_, nil), n) = ""
	  | toString' (Free (a, i), n) =
	      indent n ^ "free " ^ a ^ ": " ^
	      Int.toString (foldr op* 1 i) ^ "\n"
      in
	fun toString t = toString' (t, 0)
      end

      fun toCArr nil = ""
	| toCArr (h::t) = "[" ^ h ^ "]" ^ toCArr t

      local
	fun toCArgs nil = ""
	  | toCArgs [h] = h
	  | toCArgs (h::t) = h ^ "," ^ toCArgs t

	fun getRange id =
	  let val r = SymTab.lookup id
	  in
	    case r of
		SOME (SymTab.Int v) => v
	      | SOME _ => 0	(* these cases can't happen *)
	      | NONE => 0
	  end

	fun toC' (Null, _) = ""
	  | toC' (IntLit i, _) = IntInf.toString i ^ ".0"
	  | toC' (Var v, _) = v
	  | toC' (Fun (f, i), _) = f ^ "(" ^ toCArgs i ^ ")"
	  | toC' (Array (a, i), _) = a ^ toCArr i
	  | toC' (t as BinOp (O, l, r), n) =
	      let val s = expSize t
	      in
		(if O = Add orelse O = Sub then "(" else "") ^ toC' (l, n)
		  ^ (if s <= 5 then " " else "\n" ^ indent (n + 4))
		  ^ toStringOp O ^ " " ^ toC' (r, n)
		  ^ (if O = Add orelse O = Sub then ")" else "")
	      end
	  | toC' (Assign (l, r), n) =
	      indent n ^ toC' (l, n) ^ " = " ^ toC' (r, n) ^ ";\n"
	  | toC' (OpAssign (O, l, r), n) =
	      indent n ^ toC' (l, n) ^ " " ^ toStringOp O ^ "= "
		^ toC' (r, n) ^ ";\n"
	  | toC' (For (nil, b), n) = toC' (b, n)
	  | toC' (For (i, b), n) =
	      (* A For loop has at least one index, so `hd i' is safe. *)
	      indent n ^ "for (int " ^ hd i ^ " = 0; " ^ hd i ^ " < "
				 ^ Int.toString (getRange (hd i)) ^ "; "
				 ^ hd i ^ "++) {\n"
		^ (if length i = 1 then toC' (b, n + 4)
		   else toC' (For (tl i, b), n + 4))
		^ indent n ^ "}\n"
	  | toC' (Seq l, n) =
	      concat (map (fn t => toC' (t, n)) l)
	  | toC' (Write (w, i, l), n) =
	      indent n ^ w ^ "(" ^ toCArgs (i @ [toC' (l, n + 4)]) ^ ");\n"
	  | toC' (Return l, n) =
	      indent n ^ "return " ^ toC' (l, n + 4) ^ ";\n"
	  | toC' (VarDecl (a, i), n) =
	      indent n ^ "extern double " ^ a
		^ (if i = nil then ";\n"
		   else toCArr (map Int.toString i) ^ ";\n")
	  | toC' (FunDecl (t, f, i), n) =
	      indent n ^ "extern " ^ t ^ " " ^ f ^ "("
		(* this is a hack, if t="void", this is a writer function *)
		^ (if t = "void" then
		     toCArgs ((map (fn x => "int") i) @ ["double"])
		   else toCArgs (map (fn x => "double") i))
		^ ");\n"
	  | toC' (TempDecl (a, i), n) =
	      indent n ^ "double"
		^ (if i = nil then " " ^ a ^ ";\n"
		   else if Solution.setCostModelAllocation NONE then
		     if length i = 1 then "* " ^ a ^ ";\n"
		     else " (*" ^ a ^ ")" ^
			  toCArr (map Int.toString (tl i)) ^ ";\n"
		   else " " ^ a ^ toCArr (map Int.toString i) ^ ";\n")
	  | toC' (ResultDecl (a, i), n) =
	      indent n ^ "double"
		^ (if i = nil then " " ^ a
		   else if length i = 1 then "* " ^ a
		   else " (*" ^ a ^ ")" ^ toCArr (map Int.toString (tl i)))
		^ (if i = nil orelse Solution.setCostModelAllocation NONE then
		     ";\n"
		   else " = new double" ^ toCArr (map Int.toString i)^";\n")
	  | toC' (Function (r, f, t), n) =
	      "\n" ^ indent n 
		(* FIX: this is a hack, produce proper type representations *)
		^ (if r = "double" orelse r = "double*" orelse r = "void"
		   then r ^ " "
		   else "double (*")
		^ f ^ "()"
		^ (if r = "double" orelse r = "double*" orelse r = "void"
		   then ""
		   else ")" ^ r)
		^ " {\n"
		^ toC' (t, n + 4)
		^ indent n ^ "}\n"
	  | toC' (Alloc (_, nil), _) = ""
	  | toC' (Alloc (a, i), n) =
	      indent n ^ a ^ " = new double"
	      ^ toCArr (map Int.toString i) ^ ";\n"
	  | toC' (Free (_, nil), n) = ""
	  | toC' (Free (a, _), n) =
	      indent n ^ "delete [] " ^ a ^ ";\n"
      in
	fun toC t = toC' (t, 0)
      end
    end

    local
      (* counter for temporary variables *)
      val tempCnt = ref 0;

      val zero = IntInf.fromInt 0

      (* generate a temporary variable name and insert it into the symtab,
       * if b is true, mark the temporary as result variable *)
      fun genTempName (false, typ) =
	    let
	      val name = "t" ^ Int.toString (!tempCnt)
	    in
	      tempCnt := !tempCnt + 1;
	      SymTab.insert (name, SymTab.Temp typ);
	      name
	    end
        | genTempName (true, typ) =
	    let
	      val rname = SymTab.getResultDecl ()
	      val name = case rname of
			     SOME (a, i) => a
			   | NONE =>
			       let val n = "t" ^ Int.toString (!tempCnt)
			       in
				 tempCnt := !tempCnt + 1;
				 n
			       end
	    in
	      SymTab.insert (name, SymTab.Result typ);
	      name
	    end

      (* construct the name of the generator function for array a *)
      fun buildGeneratorCall (a, i) = Fun (a, i)

      (* build a variable or array reference *)
      fun buildRef (a, nil) = Var a
	| buildRef (a, i) = Array (a, i)

      (* build a temporary variable reference *)
      fun buildTempRef (false, i) =
	    buildRef (genTempName (false, SymTab.getRanges i), i)
	| buildTempRef (true, i) =
	    let
	      val rname = SymTab.getResultDecl ()
	      val idx = case rname of SOME (_, i) => i | NONE => i
	    in
	      buildRef (genTempName (true, SymTab.getRanges idx), idx)
	    end

      (* build an operator out of an ExpTree operator *)
      fun buildOp (ExpTree.Add) = Add
	| buildOp (ExpTree.Sub) = Sub
	| buildOp (ExpTree.Mul) = Mul
	| buildOp (ExpTree.Div) = Div

      (* build a loop if the set of indices is non-null *)
      fun buildLoop (nil, b) = b
	| buildLoop (i, b) = For (i, b)

      (* build allocation for a given reference t *)
      fun buildAllocator (t as Var a) = Alloc (a, nil)
	| buildAllocator (t as Array (a, i)) = Alloc (a, SymTab.getRanges i)
	| buildAllocator _ = raise Solution.IllegalExpressionTree

      (* build initializer for a given reference t *)
      fun buildInitializer (t as Var _) =
	    Assign (t, IntLit (IntInf.fromInt 0))
	| buildInitializer (t as Array (a, i)) =
	    For (i, Assign (t, IntLit (IntInf.fromInt 0)))
	| buildInitializer _ = raise Solution.IllegalExpressionTree

      (* build deallocation for a given reference t *)
      fun buildFree (t as Var a) = Free (a, nil)
	| buildFree (t as Array (a, i)) = Free (a, SymTab.getRanges i)
	| buildFree _ = raise Solution.IllegalExpressionTree

      (* subset relations for fusion *)
      fun ge (f', f) = IndexSet.subset (IndexSetSeq.set f, IndexSetSeq.set f')
      fun eq (f', f) = IndexSet.eq (IndexSetSeq.set f', IndexSetSeq.set f)
      fun lt (f', f) = ge (f, f') andalso not (eq (f', f))

      fun mergeFrags (nil, nil) = nil
	| mergeFrags (fl1, nil) = fl1
	| mergeFrags (nil, fl2) = fl2
	| mergeFrags (fl1 as ((f1,hl1,c1)::t1), fl2 as ((f2,hl2,c2)::t2)) =
	    let
	      val i1 = IndexSetSeq.set f1
	      val i2 = IndexSetSeq.set f2
	    in
	      if (i1 = i2 andalso 
		  (not (Solution.setCostModelAllocation NONE)
		   orelse IntInf.>= (hl1, hl2)))
		 orelse IndexSet.subset (i1, i2)
	      then (f1, hl1, c1) :: (mergeFrags (t1, fl2))
	      else (f2, hl2, c2) :: (mergeFrags (fl1, t2))
	    end

      (* genCode (f, h, frags, free)
       *   frags contains code fragments with fusion ge f;
       *   free contains deallocation fragments with fusion gt f;
       *   h contains the nesting of local loops that need to be generated;
       *   genCode generates the loops in h to include the fragments from
       *   frags and free, which results in a single fragment with fusion f
       *)
      fun genCode (_, _, nil, nil) = Null
        | genCode (f, h, nil, ffrags as (f'', _, c'')::t'') =
	    (* Sanity check: There can't be a deallocation statement
	     * with a larger fusion than the current fusion without
	     * a corresponding allocation statement in a code fragment.
	     *)
	    raise MalformedFusionTree (f, ffrags)
	| genCode (f, h, frags as (f', _, c')::t', nil) =
	    if f' = f then Seq [c', genCode (f, h, t', nil)]
	    else if h = nil then
	      (* Sanity check: f' contains indices that are missing in f@h.
	       * This case can't happen for correctly formed fusion trees.
	       * The missing loops could be inserted as follows,
	       *   Seq [buildLoop (IndexSet.difference (IndexSetSeq.set f',
	       *					IndexSetSeq.set f),
	       *		   c),
	       *	genCode (f, h, t, free)]
	       * but the generated result would have missing indices, too.
	       * Since we can't repair it properly, we raise an exception.
	       *)
	      raise MalformedFusionTree (f, frags)
	    else For (hd h, genCode (f@[hd h], tl h, frags, nil))
	| genCode (f, h, frags as (f',_,c')::t', ffrags as (f'',_,c'')::t'') =
	    if f' = f andalso f'' = f then
	      Seq [c', genCode (f, h, t', t''), c'']
	    else if f' = f then Seq [c', genCode (f, h, t', ffrags)]
	    else if f'' = f then
	      (* Sanity check: There can't be a deallocation statement
	       * with a larger fusion than the current fusion without
	       * a corresponding allocation statement in a code fragment.
	       *)
	      raise MalformedFusionTree (f, ffrags)
	    else if h = nil then
	      (* Sanity check: f' contains indices that are missing in f@h.
	       * This case can't happen for correctly formed fusion trees.
	       * The missing loops could be inserted as follows,
	       *   Seq [buildLoop (IndexSet.difference (IndexSetSeq.set f',
	       *					IndexSetSeq.set f),
	       *		   c),
	       *	genCode (f, h, t, ffrags)]
	       * but the generated result would have missing indices, too.
	       * Since we can't repair it properly, we raise an exception.
	       *)
	       raise MalformedFusionTree (f, frags)
	    else For (hd h, genCode (f@[hd h], tl h, frags, ffrags))

      (* genFreeCode (rv, ffrags)
       *   rv is the result of SymTab.getResult ();
       *   ffrags is the list of free fragments from buildSyntaxTree';
       *   generates the code for free statements with an empty fusion;
       *   free statements with a non-empty fusion have been generated in
       *   buildSyntaxTree'; the remaining free statements may contain a
       *   statement for deallocating the result, which must be removed;
       *)
      fun genFreeCode (_, nil) = Null
	| genFreeCode (NONE, (_, _, c) :: t) = Seq[c, genFreeCode(NONE, t)]
	| genFreeCode (rv as SOME (a, _), ffrags as (_, _, c) :: t) =
	    case c of
		Free (a', _) => if a = a' then genFreeCode(rv, t)
				else Seq[c, genFreeCode(rv, t)]
	      | _ => raise MalformedFusionTree ([], ffrags)

(*
      fun buildReturnVal (Array (a, i)) = Var a
	| buildReturnVal (t as _) = t
*)

      (* buildSyntaxTree' (F, f', b, (wn, wa))
       *   builds a syntax tree for a given fusion tree F and outer fusion f',
       *   the flag b is true iff F is the root of the fusion tree,
       *   the string wn is the C function name for the write function,
       *   the string list wa are index arguments for the write funciton;
       *   returns a triple (val, [(fus,code)], [(fus,code)]), where the first
       *   element is an expression for accessing the value of the tree, the
       *   second element is a list of pairs of a fusion and a code fragment
       *   for allocating and computing subtrees, and the third element is
       *   a list of pairs of fusion and code fragment for deallocating
       *   tree nodes
       *)
      fun buildSyntaxTree' (FusionTree.Leaf (_,_,_,t as ExpTree.Array (a,_)),
			    _, _, _) =
	    (* FIXME: don't fuse with unnecessary recomputation loops *)
	    (buildRef (a, ExpTree.indexList t), [], [])
	| buildSyntaxTree' (FusionTree.Leaf (f, h, (hi,lo,_),
					     t as ExpTree.Fun (a,_)),
			    f', b, _) =
	    let
	      val g = buildGeneratorCall (a, ExpTree.indexList t)
	    in
	      if null h andalso IndexSetSeq.set f = IndexSetSeq.set f' then
		(g, [], [])
	      else
		let
		  val i = IndexSetSeq.set h
		  val r = buildTempRef (b, i)
		  val loop = buildLoop (i, Assign (r, g))
		  val code = if Solution.setCostModelAllocation NONE then
			       Seq [buildAllocator r, loop]
			     else loop
		  val free = if Solution.setCostModelAllocation NONE then
			       [(f, zero, buildFree r)]
			     else []
		in
	          (r, [(f, IntInf.-(hi, lo), code)], free)
		end
	    end
	| buildSyntaxTree' (FusionTree.Leaf (_, _, _, t as ExpTree.IntLit i),
			    _, _, _) =
	    (IntLit i, [], [])
	| buildSyntaxTree' (FusionTree.Leaf _, _, _, _) =
	    raise Solution.IllegalExpressionTree
	| buildSyntaxTree' (FusionTree.Unary (f, h, (hi,lo,_),
					      t as ExpTree.Output _, l),
			    _, b, (wn, wa))=
	    let
	      val f_h = f @ h	(* should this be IndexSetSeq.extend(...)? *)
	      val (lr, lfrags, lf) = buildSyntaxTree' (l, f_h, false, (wn, wa))
	      val wCode = Write (wn, wa (* ExpTree.indices t *), lr)
	      val r = if b then Null else lr
	      val ltfrags = List.filter (fn x => lt (#1 x, f)) lfrags
	      val gefrags = List.filter (fn x => ge (#1 x, f)) lfrags
	      val ltfree = List.filter (fn x => lt (#1 x, f)) lf
	      val gefree = List.filter (fn x => ge (#1 x, f)) lf
	      val frags = mergeFrags (gefrags, [(f_h, zero, wCode)])
	    in
	      (r, mergeFrags (ltfrags, [(f, IntInf.-(hi,lo),
					 genCode (f, h, frags, gefree))]),
	       ltfree)
	    end
	| buildSyntaxTree' (FusionTree.Unary (f, h, (hi,lo,_),
					      t as ExpTree.Cast _, l),
			    _, b, w)=
	    let
	      val f_h = f @ h
	      val (lr, lfrags, lf) = buildSyntaxTree' (l, f_h, false, w)
	      val ltfrags = List.filter (fn x => lt (#1 x, f)) lfrags
	      val gefrags = List.filter (fn x => ge (#1 x, f)) lfrags
	      val ltfree = List.filter (fn x => lt (#1 x, f)) lf
	      val gefree = List.filter (fn x => ge (#1 x, f)) lf
	    in
	      if null h then
		(lr, mergeFrags (ltfrags, [(f, zero,
					    genCode (f, h, gefrags, gefree))]),
		 ltfree)
	      else
		let
	          val r = buildTempRef (b, IndexSetSeq.set h)
	          val rCode = Assign (r, lr)
		  val frags = if Solution.setCostModelAllocation NONE then
				mergeFrags (gefrags,
					    [(f, zero, buildAllocator r),
					     (f_h, zero, rCode)])
			      else
				mergeFrags (gefrags, [(f_h, zero, rCode)])
		  val free = if Solution.setCostModelAllocation NONE then
			       [(f, zero, buildFree r)]
			     else []
		in
		  (r, mergeFrags (ltfrags, [(f, IntInf.-(hi,lo),
					     genCode (f, h, frags, gefree))]),
		   mergeFrags (ltfree, free))
		end
	    end
	| buildSyntaxTree' (FusionTree.Unary (f, h, (hi,lo,_),
					      t as ExpTree.Sum _, l),
			    _, b, w)=
	    let
	      val f_h = f @ h
	      val (lr, lfrags, lf) = buildSyntaxTree' (l, f_h, false, w)
	      val i = IndexSet.intersect (IndexSetSeq.set h, ExpTree.fusible t)
	      val r = buildTempRef (b, i)
	      val init = if Solution.setCostModelAllocation NONE then
			   Seq [buildAllocator r, buildInitializer r]
			 else
			   buildInitializer r
	      val free = if Solution.setCostModelAllocation NONE then
			   [(f, zero, buildFree r)]
			 else []
	      val rCode = OpAssign (Add, r, lr)
	      val ltfrags = List.filter (fn x => lt (#1 x, f)) lfrags
	      val gefrags = List.filter (fn x => ge (#1 x, f)) lfrags
	      val ltfree = List.filter (fn x => lt (#1 x, f)) lf
	      val gefree = List.filter (fn x => ge (#1 x, f)) lf
	      val frags = mergeFrags (gefrags, [(f, zero, init),
						(f_h, zero, rCode)])
	    in
	      (r, mergeFrags (ltfrags, [(f, IntInf.-(hi,lo),
					 genCode (f, h, frags, gefree))]),
	       mergeFrags(ltfree, free))
	    end
	| buildSyntaxTree' (FusionTree.Unary _, _, _, _) =
	    raise Solution.IllegalExpressionTree
	| buildSyntaxTree' (FusionTree.Binary (f, h, (hi,lo,_),
					       t as ExpTree.BinOp(O,_,_),
					       l, r),
			    f', b, w) =
	    let
	      val f_h = f @ h
	      val (lr, lfrags, lf) = buildSyntaxTree' (l, f_h, false, w)
	      val (rr, rfrags, rf) = buildSyntaxTree' (r, f_h, false, w)
	      val exp = BinOp (buildOp O, lr, rr)
	      val mfrags = mergeFrags (lfrags, rfrags)
	      val mfree = mergeFrags (rf, lf)
	      val ltfrags = List.filter (fn x => lt (#1 x, f)) mfrags
	      val gefrags = List.filter (fn x => ge (#1 x, f)) mfrags
	      val ltfree = List.filter (fn x => lt (#1 x, f)) mfree
	      val gefree = List.filter (fn x => ge (#1 x, f)) mfree
	    in
	      if null h andalso IndexSetSeq.set f = IndexSetSeq.set f' then
		if gefrags = nil andalso gefree = nil then
		  (exp, ltfrags, ltfree)
		else (exp, mergeFrags (ltfrags,
				       [(f,zero,genCode(f,h,gefrags,gefree))]),
		      ltfree)
	      else
		let
		  val i = IndexSetSeq.set h
		  val r = buildTempRef (b, i)
		  val rCode = Assign (r, exp)
		  val frags = if Solution.setCostModelAllocation NONE then
				mergeFrags (gefrags,
					    [(f, zero, buildAllocator r),
					     (f_h, zero, rCode)])
			      else
				mergeFrags (gefrags, [(f_h, zero, rCode)])
		  val free = if Solution.setCostModelAllocation NONE then
			       [(f, zero, buildFree r)]
			     else []
		in
		  (r, mergeFrags (ltfrags, [(f, IntInf.-(hi,lo),
					     genCode (f, h, frags, gefree))]),
		   mergeFrags (ltfree, free))
		end
	    end
	| buildSyntaxTree' (FusionTree.Binary _, _, _, _) =
	    raise Solution.IllegalExpressionTree
    in
      (* buildSyntaxTree (D, F)
       *   builds a syntax tree for definition D and fusion tree F
       *)
      fun buildSyntaxTree (D, F) =
	(tempCnt := 0;
	 SymTab.deleteTemps ();
	 let
	   val (name, writer) =
		 case D of
		     NONE => ("compute", ("write", nil))
		   | SOME (ExpTree.Array (a, i)) =>
		       let val (n, w) = ("compute_"^a, ("write_"^a, map #1 i))
		       in
		         SymTab.insert (a, SymTab.ResultDecl (map #1 i));
			 (n, w)
		       end
		   | SOME (ExpTree.Output (ExpTree.Array (a, i))) =>
		       let val (n, w) = ("compute_"^a, ("write_"^a, map #1 i))
		       in
		         SymTab.insert (#1 w, SymTab.Writer (#2 w));
			 (n, w)
		       end
		   | SOME _ => ("compute", ("write", nil)) (* can't happen *)

	   val (r, frags, ffrags) = buildSyntaxTree' (F, [], true, writer)
	   val td = Seq (map TempDecl (SymTab.getTemps ()))
	   val rv = SymTab.getResult ()
	   val free = genFreeCode (rv, ffrags)
	   val res = case rv of
			 SOME (a,i) => Return (Var a) (* buildReturnVal r *)
		       | NONE => if r = Null then Null
				 else
				   let val tmp = genTempName (true, nil)
				   in
				     Seq [Assign (buildRef (tmp, nil), r),
					  Return (Var tmp)]
				   end
	   val rv' = SymTab.getResult ()
	   val rt = case rv' of
			 SOME (_, i) => if i = nil then "double"
					else if length i = 1 then "double*"
					(* FIX: construct proper types *)
					else toCArr (map Int.toString (tl i))
		       | NONE => "void"
	   val rd = case rv' of
			 SOME v => ResultDecl v
		       | NONE => Null
	 in
	   Function (rt, name,
		     Seq [td, rd,
			  if frags = nil then Null else #3 (hd frags),
			  free,
			  res])
	 end
	)

      (* buildSyntaxTreeFile fl
       *   builds the syntax tree for multiple fusion trees
       *)
      fun buildSyntaxTreeFile fl =
	let
	  val stl = Seq (map buildSyntaxTree fl)
	  val fd = Seq (map (fn (n, i) => FunDecl ("double", n, i))
			    (SymTab.getFuns ()))
	  (* FIX: elements of the third arg of FunDecl aren't used anyway *)
	  val wd = Seq (map (fn (n, i) => FunDecl ("void", n, map size i))
			    (SymTab.getWriters ()))
	  val vd = Seq (map VarDecl (SymTab.getArrays ()))
	in
	  Seq [fd, wd, vd, stl]
	end
    end

  end
