(***************************************************************************
 *
 * Symbol Table Implementation
 *
 * Author: Gerald Baumgartner
 * Date:   27 October 2001
 *
 ***************************************************************************)

structure SymTab : SYM_TAB =
  struct
    datatype elem = Int of int
		  | Array of int list
		  | Fun of int list
		  | Temp of int list
		  | Result of int list
		  | ResultDecl of string list
		  | Writer of string list

    type symTab = (string * elem) list

    val table : symTab ref = ref nil

    (* clear ()
     *   Clears the symbol table.
     *)
    fun clear () = table := nil

    (* deleteTemps ()
     *   Deletes all temporaries from the symbol table.
     *)
    local
      fun deleteTemps' nil = nil
	| deleteTemps' ((_, Temp _) :: t) = deleteTemps' t
	| deleteTemps' ((_, Result _) :: t) = deleteTemps' t
	| deleteTemps' (h :: t) = h :: deleteTemps' t
    in
      fun deleteTemps () = table := deleteTemps' (! table)
    end

    (* insert (key, val)
     *   Insert the pair (key, val) in the table, shadowing any previous value.
     *)
    fun insert (key, value) = table := (key, value) :: !table

    (* lookup key
     *   Lookup the value associated with key in the table.
     *)
    local
      fun lookup' (key, nil) = NONE
        | lookup' (key, (k,v)::t) =
	    if key = k then SOME v else lookup' (key, t)
    in
      fun lookup key = lookup' (key, !table)
    end

    (* getArrays ()
     *   Extracts the list of arrays from the symbol table.
     *)
    local
      fun getArrays' nil = nil
	| getArrays' ((a, Array i) :: t) = (a, i) :: getArrays' t
	| getArrays' (_ :: t) = getArrays' t
    in
      fun getArrays () = rev (getArrays' (! table))
    end

    (* getFuns ()
     *   Extracts the list of functions from the symbol table.
     *)
    local
      fun getFuns' nil = nil
	| getFuns' ((f, Fun i) :: t) = (f, i) :: getFuns' t
	| getFuns' (_ :: t) = getFuns' t
    in
      fun getFuns () = rev (getFuns' (! table))
    end

    (* getTemps ()
     *   Extracts the list of temporaries from the symbol table.
     *)
    local
      fun getTemps' nil = nil
	| getTemps' ((a, Temp i) :: t) = (a, i) :: getTemps' t
	| getTemps' (_ :: t) = getTemps' t
    in
      fun getTemps () = rev (getTemps' (! table))
    end

    (* getResult ()
     *   Extract the result variable from the symbol table.
     *)
    local
      fun getResult' nil = NONE
	| getResult' ((r, Result i) :: t) = SOME (r, i)
	| getResult' (_ :: t) = getResult' t
    in
      fun getResult () = getResult' (! table)
    end

    (* getResultDecl ()
     *   Extract the result variable from the symbol table.
     *)
    local
      fun getResultDecl' nil = NONE
	| getResultDecl' ((r, ResultDecl i) :: t) = SOME (r, i)
	| getResultDecl' (_ :: t) = getResultDecl' t
    in
      fun getResultDecl () = getResultDecl' (! table)
    end

    (* getWriters ()
     *   Extracts the list of writer functions from the symbol table.
     *)
    local
      fun getWriters' nil = nil
	| getWriters' ((f, Writer i) :: t) = (f, i) :: getWriters' t
	| getWriters' (_ :: t) = getWriters' t
    in
      fun getWriters () = rev (getWriters' (! table))
    end

    (* getRange id
     *   Extracts the range of an index identifier id; reports an error
     *   if id is not an index identifier or was not declared.
     *)
    fun getRange ident =
      let val r = lookup ident
      in
	case r of
	    SOME (Int v) => v
	  | SOME _ => (print ("Error: identifier " ^ ident ^
			      " is not an index\n");
		       1)
	  | NONE   => (print ("Error: no range defined for index " ^ ident ^
			      ", using range 1\n");
		       insert (ident, Int 1);
		       1)
      end

    (* gets the types of a list of index variables *)
    fun getRanges nil = nil
      | getRanges (id::t) =
	  let val r = lookup id
	  in
	    case r of
		SOME (Int v) => v :: (getRanges t)
	      | SOME _ => (print ("Error: identifier " ^ id
				    ^ " is not an index\n");
			   1 :: (getRanges t))
	      | NONE   => (print ("Error: identifier " ^ id
				    ^ " undefined\n");
			   1 :: (getRanges t))
	  end

    val nameMangling = ref false;

    fun mangleNames NONE = !nameMangling
      | mangleNames (SOME b) =
	  let val old = !nameMangling
	  in
	    nameMangling := b;
	    old
	  end

  end
