(***************************************************************************
 *
 * Top-Level Structure
 *
 * Author: Gerald Baumgartner
 * Date:   18 October 2001
 * Date:   9 April 2010
 *
 ***************************************************************************)

structure TopLevel =
  struct

    open ExpTree
    open SolTree
    open FusionTree

(*
    structure ExpLrVals = ExpLrValsFun (
      structure Token = LrParser.Token
      structure ExpTree = ExpTree);

    structure ExpLex = ExpLex (structure Tokens = ExpLrVals.Tokens);

    structure ExpParser = Join (
      structure LrParser = LrParser
      structure ParserData = ExpLrVals.ParserData
      structure Lex = ExpLex);
*)

    structure EqLrVals = EqLrValsFun (
      structure Token = LrParser.Token
      structure IndexSet = IndexSet
      structure ExpTree = ExpTree
      structure SymTab = SymTab);

    structure EqLex = EqLex (structure Tokens = EqLrVals.Tokens);

    structure EqParser = Join (
      structure LrParser = LrParser
      structure ParserData = EqLrVals.ParserData
      structure Lex = EqLex);

    local
      fun error (errmsg, line, _) =
	print ("Parse error.  Line " ^ (Int.toString line) ^ ": "
		 ^ errmsg ^ "\n")

      fun parse' file =
	let
	  fun reader _ = TextIO.input file
	  val lexer = EqParser.makeLexer reader
	  val (tree, _) = EqParser.parse (15, lexer, error, ())
	in
	  TextIO.closeIn file;
	  tree
	end

    in
      (* parseString s
       *   Run the parser on string s.
       *)
      fun parseString string =
	parse' (TextIO.openString string)

      (* parse f
       *   Run the parser on file f.
       *)
      fun parse filename = 
	parse' (TextIO.openIn filename)
    end


(* Printing Functions *)
(* ------------------ *)

    (* pt t
     *   print an expression tree
     *)
    fun pt t = print (ExpTree.toString t)

    (* pe t
     *   print an expression tree as an expression
     *)
    fun pe t = print ((ExpTree.toExp t) ^ "\n")

    (* ps x
     *   print a solution tree
     *)
    fun ps x = print (SolTree.toString x)

    (* pf x
     *   print a fusion tree
     *)
    fun pf x = print (FusionTree.toString x)


    (* pst t
     *   compute and print a solution tree for expression tree t
     *)
    fun pst t = print (SolTree.toString (SolTree.minMemFusion t))

    (* pft t
     *   compute and print a fusion tree for expression tree t
     *)
    fun pft t = print (FusionTree.toString (FusionTree.expTreeToFusionTree t))


(* Print Pseudo Code *)
(* ----------------- *)

    (* pc t
     *   take an expression tree, compute the fusion tree, and spit out code
     *)
    fun pc t =
      let val ft = FusionTree.expTreeToFusionTree t
      in
	print (AbSyn.toString (AbSyn.buildSyntaxTree (NONE, ft)))
      end

    (* pnc t
     *   take an expression tree and spit out unfused code
     *)
    fun pnc t =
      let val ft = FusionTree.buildUnfusedFusionTree t
      in
	print (AbSyn.toString (AbSyn.buildSyntaxTree (NONE, ft)))
      end

    (* pcl dl
     *   take a list of pairs containing a name and an expression tree,
     *   compute all the fusion trees, and generate the code for an
     *   entire compilation unit
     *)
    fun pcl dl =
      let val ftl = map (fn (d, t) => (d, FusionTree.expTreeToFusionTree t))
			dl
      in
	print (AbSyn.toString (AbSyn.buildSyntaxTreeFile ftl))
      end

    (* pncl dl
     *   take a list of pairs containing a name and an expression tree,
     *   and generate unfused code for an entire compilation unit
     *)
    fun pncl dl =
      let val ftl = map (fn (d, t) => (d, FusionTree.buildUnfusedFusionTree t))
			dl
      in
	print (AbSyn.toString (AbSyn.buildSyntaxTreeFile ftl))
      end


(* Print Debug Information *)
(*------------------------ *)

    local
      fun et2ft t =
	let val st = SolTree.minMemFusion t
	in
	  print ("Solution Tree:\n\n" ^ SolTree.toString st ^ "\n\n");
	  let val ft = FusionTree.buildFusionTree st
	  in
	    print ("Fusion Tree:\n\n" ^ toString ft ^ "\n\n");
	    ft
	  end
	end

      fun comp NONE = ()
	| comp (SOME d) = print ("Compiling " ^ (ExpTree.toExp d) ^ ":\n\n");
    in

      (* pd t
       *   take an expression tree, compute the fusion tree, and spit out code;
       *   print solution tree and fusion tree as debug information
       *)
      fun pd t =
	let val ast = AbSyn.buildSyntaxTree (NONE, et2ft t)
	in
	  print ("Code:\n\n" ^ AbSyn.toString ast)
	end
	handle NoSolution =>
	  (print "No solution found.\n")
	handle AbSyn.MalformedFusionTree (f, frags) =>
	  (
	   print "Malformed Fusion Tree in AbSyn.genCode\n\n";
	   print ("f = " ^ IndexSetSeq.toString f ^ ";  h = nil\n");
	   print "frags:\n";
	   map (fn (f, _, c) =>
		   (print ("\n// f = " ^ IndexSetSeq.toString f ^ "\n");
		    print (AbSyn.toString c)))
	       frags;
	   print "\n"
	  )

      (* pdl dl
       *   take a list of pairs containing a name and an expression tree,
       *   compute all the fusion trees, and generate the code for an
       *   entire compilation unit;
       *   print solution tree and fusion tree as debug information
       *)
      fun pdl dl =
	let val ftl = map (fn (d, t) => (d, (comp d; et2ft t)))
			  dl
	in
	  print ("Code:\n\n" ^ AbSyn.toString (AbSyn.buildSyntaxTreeFile ftl))
	end
	handle NoSolution =>
	  (print "No solution found.\n")
        handle AbSyn.MalformedFusionTree (f, frags) =>
	  (
	   print "Malformed Fusion Tree in AbSyn.genCode\n\n";
	   print ("f = " ^ IndexSetSeq.toString f ^ ";  h = nil\n");
	   print "frags:\n";
	   map (fn (f, _, c) =>
		   (print ("\n// f = " ^ IndexSetSeq.toString f ^ "\n");
		    print (AbSyn.toString c)))
	       frags;
	   print "\n"
	  )
    end


(* Print Timing Results *)
(* -------------------- *)

    (* N
     *   the number of time measurements to average
     *)
    val N = ref 10

    (* STT
     *   whether space-time tradeoffs should be timed as well
     *)
    val STT = ref true

    local
      fun et2ft' t =
	(Solution.clearMemoryLimit ();
	 let
	   val timer = Timer.startCPUTimer()
	   val st = SolTree.minMemFusion t
	   val t1 = #usr (Timer.checkCPUTimer timer)
           val Solution.S (_, (c, _, _, _), _, _, _) = getSoln st
	   val ft = FusionTree.buildFusionTree st
	   val t2 = #usr (Timer.checkCPUTimer timer)
	 in
	   if (!STT) then
	     (Solution.setMemoryLimit c;
	      let
		val dummy = SolTree.spaceTimeTradeoffs t
		val t3 = #usr (Timer.checkCPUTimer timer)
	      in
		(t1, t2, t3, ft)
	      end
	     )
	   else
	     (t1, t2, t2, ft)
	 end
	)

      fun net2ft (n, s1, s2, s3, t) =
	let
	  val (t1, t2, t3, ft) = et2ft' t
	in
	  if n = 1 then
	    (Time.+ (s1, t1), Time.+ (s2, t2), Time.+ (s3, t3), ft)
	  else
	    net2ft (n-1, Time.+ (s1, t1), Time.+ (s2, t2), Time.+ (s3, t3), t)
	end

      fun et2ft t =
	let
	  val (t1, t2, t3, ft) =
	      net2ft (!N, Time.zeroTime, Time.zeroTime, Time.zeroTime, t)
	  val avgT1 = (Time.toReal t1) / (LargeReal.fromInt (!N))
	  val avgT2 = (Time.toReal (Time.- (t2, t1))) / (LargeReal.fromInt(!N))
	  val avgT3 = (Time.toReal (Time.- (t3, t2))) / (LargeReal.fromInt(!N))
	in
	  print ("Build solution tree:  " ^ (LargeReal.toString avgT1) ^ "\n");
	  print ("Build fusion tree:    " ^ (LargeReal.toString avgT2) ^ "\n");
	  if (!STT) then
	    print ("Space-time Tradeoffs: " ^ (LargeReal.toString avgT3)^"\n")
	  else
	    ();
	  ft
	end

      fun comp NONE = ()
	| comp (SOME d) = print ("Timing " ^ (ExpTree.toExp d) ^ ":\n\n");
    in

      (* ptc t
       *   take an expression tree, compute the fusion tree, and spit out code;
       *   print timing information for all the traversals
       *)
    
      fun ptc t =
	let
	  val ft = et2ft t
	  val ast = AbSyn.buildSyntaxTree (NONE, ft)
	in
	  print (AbSyn.toString ast)
	end

      (* ptcl dl
       *   take a list of pairs containing a name and an expression tree,
       *   compute all the fusion trees, and generate the code for an
       *   entire compilation unit;
       *   print timing information for all the traversals
       *)

      fun ptcl dl =
	let
	  val ftl = map (fn (d, t) => (d, (comp d; et2ft t)))
			dl
	in
	  print (AbSyn.toString (AbSyn.buildSyntaxTreeFile ftl))
	end

    end


(* Print Space-Time Tradeoffs *)
(* -------------------------- *)

    local
      fun pstl' (def, sttl) =
	map (fn st =>
	      let
		val Solution.S (_, (c, _, _, _), _, rc, _) = getSoln st
		val ft = FusionTree.buildFusionTree st
		val ast = AbSyn.buildSyntaxTree (def, ft)
	      in
	        print "\n\n";
		print ("Space: " ^ IntInf.toString c ^ "\n");
		print ("Time:  " ^ IntInf.toString rc ^ "\n\n");
(*
		print ("Solution Tree:\n\n" ^ SolTree.toString st ^ "\n\n");
		print ("Fusion Tree:\n\n" ^ toString ft ^ "\n\n");
 *)
		print (AbSyn.toString ast)
	      end)
	    sttl
    in
      (* pstl dl
       *   take a list of pairs containing a name and an expression tree,
       *   compute all the fusion trees, and generate the code for an
       *   entire compilation unit;
       *   print all space-time tradeoff solutions with their cost
       *)
      fun pstl dl =
	(map (fn (d, t) =>
	       (print "Space-Time Tradeoffs:";
		Solution.clearMemoryLimit ();
		let
		  val st = SolTree.minMemFusion t
		  val Solution.S (_, (c, _, _, _), _, _, _) = getSoln st
		in
		  Solution.setMemoryLimit c;
	          pstl' (d, SolTree.spaceTimeTradeoffs t)
		end;
		print "\n\n"
	       ))
	     dl;
	 ())
    end


(* Print C Code *)
(* ------------ *)

    (* pC t
     *   take an expression tree, compute the fusion tree, and spit out C code
     *)
    fun pC t =
      let val ft = FusionTree.expTreeToFusionTree t
      in
	print (AbSyn.toC (AbSyn.buildSyntaxTree (NONE, ft)))
      end

    (* pnC t
     *   take an expression tree and spit out unfused C code
     *)
    fun pnC t =
      let val ft = FusionTree.buildUnfusedFusionTree t
      in
	print (AbSyn.toC (AbSyn.buildSyntaxTree (NONE, ft)))
      end

    (* pCl dl
     *   take a list of pairs containing a name and an expression tree,
     *   compute all the fusion trees, and generate the code for an
     *   entire compilation unit as C code
     *)
    fun pCl dl =
      let val ftl = map (fn (d, t) => (d, FusionTree.expTreeToFusionTree t))
			dl
      in
	print (AbSyn.toC (AbSyn.buildSyntaxTreeFile ftl))
      end

    (* pnCl dl
     *   take a list of pairs containing a name and an expression tree,
     *   and generate unfused code for an entire compilation unit as C code
     *)
    fun pnCl dl =
      let val ftl = map (fn (d, t) => (d, FusionTree.buildUnfusedFusionTree t))
			dl
      in
	print (AbSyn.toC (AbSyn.buildSyntaxTreeFile ftl))
      end


(* Main *)
(* ---- *)

    (* The main functions are:
     *   val pt   : expTree -> unit	(* print an expression tree *)
     *   val pe   : expTree -> unit	(* print an expression *)
     *   val pft  : expTree -> unit	(* print the fusion tree *)
     *   val pst  : expTree -> unit	(* print the solution tree *)
     *
     *   val pc   : expTree -> unit	(* print the code *)
     *   val pd   : expTree -> unit	(* print the code in debug mode *)
     *   val ptc  : expTree -> unit	(* print execution times and code *)
     *   val pst  : expTree -> unit	(* print space-time tradeoffs *)
     *   val pC   : expTree -> unit	(* print the C code *)
     *
     * And corresponding functions for lists of pairs of LHS and expTree:
     *   val pcl  : (expTree option * expTree) list -> unit
     *   val pdl  : (expTree option * expTree) list -> unit
     *   val ptcl : (expTree option * expTree) list -> unit
     *   val pstl : (expTree option * expTree) list -> unit
     *   val pCl  : (expTree option * expTree) list -> unit
     *)

    (* main' f
     *   Parse file f and print the code according to options.
     *)
    fun main' args = 
      (SymTab.clear ();
       if length args = 1 then pcl (parse (hd args))
       else if length args = 2 then
	 let
	   val opt = explode (hd args)
	   val file = hd (tl args)
	 in
	   Solution.setCostModelDetails
	     (SOME (not (List.exists (fn x => x = #"b") opt)));
	   Solution.setCostModelAllocation
	     (SOME (List.exists (fn x => x = #"a") opt));
	   SymTab.mangleNames (SOME (List.exists (fn x => x = #"m") opt));
	   if List.exists (fn x => x = #"n") opt andalso
	      List.exists (fn x => x = #"c") opt then pnCl (parse file)
	   else if List.exists (fn x => x = #"n") opt then pncl (parse file)
	   else if List.exists (fn x => x = #"c") opt then pCl (parse file)
	   else if List.exists (fn x => x = #"d") opt then pdl (parse file)
	   else if List.exists (fn x => x = #"t") opt then ptcl (parse file)
	   else if List.exists (fn x => x = #"s") opt then pstl (parse file)
	   else pcl (parse file)
	 end
       else
	 print ("Usage: minmem [-abcdmst] file\n\n" ^
		"  -a\t\tuse dynamic memory allocation cost model\n" ^
		"  -b\t\tuse basic static memory allocation cost model\n" ^
		"  -c\t\tgenerate C code\n" ^
		"  -d\t\tprint debug information\n" ^
		"  -m\t\tgenerate mangled array and function names\n" ^
		"  -n\t\tgenerate unfused code\n" ^
		"  -s\t\toverride limit, print all space-time tradeoffs\n" ^
		"  -t\t\tprint execution times\n")
      )
      handle Io => print "Error: exception was raised, e.g., file not found\n"

    (* main ()
     *   Get filename from command line, parse file, print code
     *   according to the command line options.
     *)
    fun main (name, args) =
      (main' args;
       OS.Process.success)
  end
